/* -*- c++ -*- */
/*
 * Copyright 2021 Ralf Gorholt.
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifndef INCLUDED_DL5EU_TPS_DECODER_IMPL_H
#define INCLUDED_DL5EU_TPS_DECODER_IMPL_H

#include <dl5eu/tps_decoder.h>

const int SYMBOLS_PER_FRAME = 68;
const int FRAMES_PER_SUPERFRAME = 4;

const int CONTINUAL_PILOT_SIZE_2K = 45;
const int TPS_PILOT_SIZE_2K = 17;
const int CONTINUAL_PILOT_SIZE_8K = 177;
const int TPS_PILOT_SIZE_8K = 68;

namespace gr {
namespace dl5eu {

class tps_decoder_impl : public tps_decoder
{
private:
    // These types are used while decoding the TPS carriers
    typedef enum _modulation_scheme { QPSK, QAM16, QAM64, M_UNUSED = 7 } modulation_scheme_t;
    typedef enum _convolutional_code { RATE_1_2, RATE_2_3, RATE_3_4, RATE_5_6, RATE_7_8, C_UNUSED = 7 } convolutional_code_t;
    typedef enum _guard_interval { GUARD_1_32, GUARD_1_16, GUARD_1_8, GUARD_1_4, G_UNUSED = 7 } guard_interval_t;

    static const int d_symbols_per_frame;
    static const int d_frames_per_superframe;

    // Continual pilot carriers info
    static const int d_cpilot_carriers_size_2k;
    static const int d_cpilot_carriers_2k[];
    static const int d_cpilot_carriers_size_8k;
    static const int d_cpilot_carriers_8k[];

    // TPS carriers info
    static const int d_tps_carriers_size_2k;
    static const int d_tps_carriers_2k[];
    static const int d_tps_carriers_size_8k;
    static const int d_tps_carriers_8k[];

    // TPS sync data
    static const int d_tps_sync_size;
    static const int d_tps_sync_even[];
    static const int d_tps_sync_odd[];

    int d_cpilot_carriers_size;
    const int* d_cpilot_carriers;
    int d_tps_carriers_size;
    const int *d_tps_carriers;

    // Number of data carriers
    static const int d_data_carriers_2k;
    static const int d_data_carriers_8k;

    // Number of active carriers
    int d_active_carriers;

    // The total number of data carriers
    int d_total_data_carriers;

    // Number of symbols per frame
    //static const int d_symbols_per_frame;

    //static const int d_tps_sync_size;

    // Keeps the rcv TPS data, is a FIFO
    // TODO circular_buffers are superior to deques.
    // We should migrate them all.
    std::deque<char> d_rcv_tps_data;
    // Keeps the TPS sync sequence
    std::deque<char> d_tps_sync_evenv;
    std::deque<char> d_tps_sync_oddv;

    //static const int d_tps_sync_even[];
    //static const int d_tps_sync_odd[];

    // TPS parity check matrix
    static const char d_h[];

    // Whether or not the decoded TPS parameters should be printed.
    bool d_print_params;

    // Keep TPS carriers values from previous symbol
    gr_complex *d_prev_tps_symbol;

    // See if the symbol index is known
    int d_symbol_index_known;

    // Indicates whether the last symbol was a frame end
    bool d_frame_end;
    // Indicates whther a re-sync should be signaled downstream
    bool d_resync;

    // Symbol Index
    int d_symbol_index;

    int number_symbol;

    // how many symbols ago we saw a complete tps
    int d_since_last_tps;

    // A list of the data carriers for the current configuration (mode), one
    // list per symbol.
    int *d_data_carriers;

    /*!
     * This method checks whether or not the BCH code in d_rcv_tps_data is correct.
     */
    int verify_bch_code(std::deque<char> data);

    void print_modulation_scheme(modulation_scheme_t modulation_scheme);
    void print_convolutional_code(convolutional_code_t convolutional_code);
    void print_guard_interval(guard_interval_t guard_interval);

    /*!
     * Prints the decoded TPS data in d_rcv_tps_data in a somewhat nice and human-readable format.
     */
    void tps_print();

    /*!
     * \brief The method decodes the TPS carriers of a given symbol, thus appending a new bit
     * to those received so far. If 68 were received since the last complete tps, we check whether
     * the sync word if present at the beginning, and if it passes the BCH parity code.
     *
     * Our present implementation simply performs the differential demodulation (TPS is transmitted in
     * DBPSK) for every carrier, and then performs a majority vote between all carriers to verify which
     * bit was sent.
     */
    int process_tps_data(const gr_complex *in);

    /*!
     * Sets several parameters used throughout the rest of the methods
     */
    void set_carriers_parameters(int payload);

    /*!
     *  it constructs the d_data_carriers list, which contains the list of carriers that have actual DTV data.
     */
    void construct_data_carriers_list();

public:
    tps_decoder_impl(int mode, bool print_params);
    ~tps_decoder_impl();

    // Where all the action really happens
    void forecast(int noutput_items, gr_vector_int &ninput_items_required);

    int general_work(int noutput_items,
                     gr_vector_int &ninput_items,
                     gr_vector_const_void_star &input_items,
                     gr_vector_void_star &output_items);
};

} // namespace dl5eu
} // namespace gr

#endif /* INCLUDED_DL5EU_TPS_DECODER_IMPL_H */
